@file:JvmName("_IntConversions_Sign")
@file:Suppress("NOTHING_TO_INLINE")

package ru.landrafhomyak.kotlin.utilities

import kotlin.contracts.InvocationKind
import kotlin.contracts.contract
import kotlin.jvm.JvmName


/********************************************* Reinterpret sign casts ****************************************/


/**
 * Converts this [Byte] value to [UByte].
 *
 * If this value is positive, the resulting `UByte` value represents the same numerical value as this `Byte`.
 *
 * The resulting `UByte` value has the same binary representation as this `Byte` value.
 */
@JvmName("asUByte")
public inline fun Byte.asUByte(): UByte = this.toUByte()


/**
 * Converts this [UByte] value to [Byte].
 *
 * If this value is less than or equals to [Byte.MAX_VALUE], the resulting `Byte` value represents
 * the same numerical value as this `UByte`. Otherwise the result is negative.
 *
 * The resulting `Byte` value has the same binary representation as this `UByte` value.
 */
@JvmName("asByte")
public inline fun UByte.asByte(): Byte = this.toByte()

/**
 * Converts this [Short] value to [UShort].
 *
 * If this value is positive, the resulting `UShort` value represents the same numerical value as this `Short`.
 *
 * The resulting `UShort` value has the same binary representation as this `Short` value.
 */
@JvmName("asUShort")
public inline fun Short.asUShort(): UShort = this.toUShort()


/**
 * Converts this [UShort] value to [Short].
 *
 * If this value is less than or equals to [Short.MAX_VALUE], the resulting `Short` value represents
 * the same numerical value as this `UShort`. Otherwise the result is negative.
 *
 * The resulting `Short` value has the same binary representation as this `UShort` value.
 */
@JvmName("asShort")
public inline fun UShort.asShort(): Short = this.toShort()

/**
 * Converts this [Int] value to [UInt].
 *
 * If this value is positive, the resulting `UInt` value represents the same numerical value as this `Int`.
 *
 * The resulting `UInt` value has the same binary representation as this `Int` value.
 */
@JvmName("asUInt")
public inline fun Int.asUInt(): UInt = this.toUInt()

/**
 * Converts this [UInt] value to [Int].
 *
 * If this value is less than or equals to [Int.MAX_VALUE], the resulting `Int` value represents
 * the same numerical value as this `UInt`. Otherwise the result is negative.
 *
 * The resulting `Int` value has the same binary representation as this `UInt` value.
 */
@JvmName("asInt")
public inline fun UInt.asInt(): Int = this.toInt()


/**
 * Converts this [Long] value to [ULong].
 *
 * If this value is positive, the resulting `ULong` value represents the same numerical value as this `Long`.
 *
 * The resulting `ULong` value has the same binary representation as this `Long` value.
 */
@JvmName("asULong")
public inline fun Long.asULong(): ULong = this.toULong()

/**
 * Converts this [ULong] value to [Long].
 *
 * If this value is less than or equals to [Long.MAX_VALUE], the resulting `Long` value represents
 * the same numerical value as this `ULong`. Otherwise the result is negative.
 *
 * The resulting `Long` value has the same binary representation as this `ULong` value.
 */
@JvmName("asLong")
public inline fun ULong.asLong(): Long = this.toLong()


/********************************************* Safe sign casts (or null) ****************************************/


/**
 * Converts this [Byte] value to [UByte].
 *
 * If this value is positive, the resulting [UByte] value represents the same numerical value as this [Byte].
 * Otherwise `null` returned.
 */

@JvmName("toUByteOrNull")
public inline fun Byte.toUByteOrNull(): UByte? = if (this < 0) null else this.asUByte()


/**
 * Converts this [UByte] value to [Byte].
 *
 * If this value is less than or equals to [Byte.MAX_VALUE], the resulting [Byte] value represents
 * the same numerical value as this [UByte].
 * Otherwise `null` returned.
 */
@JvmName("toByteOrNull")
public inline fun UByte.toByteOrNull(): Byte? = if (this >= Byte.MAX_VALUE.asUByte()) null else this.asByte()

/**
 * Converts this [Short] value to [UShort].
 *
 * If this value is positive, the resulting [UShort] value represents the same numerical value as this [Short].
 * Otherwise `null` returned.
 */
@JvmName("toUShortOrNull")
public inline fun Short.toUShortOrNull(): UShort? = if (this < 0) null else this.asUShort()


/**
 * Converts this [UShort] value to [Short].
 *
 * If this value is less than or equals to [Short.MAX_VALUE], the resulting [Short] value represents
 * the same numerical value as this [UShort].
 * Otherwise `null` returned.
 */
@JvmName("toShortOrNull")
public inline fun UShort.toShortOrNull(): Short? = if (this >= Short.MAX_VALUE.asUShort()) null else this.asShort()

/**
 * Converts this [Int] value to [UInt].
 *
 * If this value is positive, the resulting [UInt] value represents the same numerical value as this [Int].
 * Otherwise `null` returned.
 */
@JvmName("toUIntOrNull")
public inline fun Int.toUIntOrNull(): UInt? = if (this < 0) null else this.asUInt()

/**
 * Converts this [UInt] value to [Int].
 *
 * If this value is less than or equals to [Int.MAX_VALUE], the resulting [Int] value represents
 * the same numerical value as this [UInt].
 * Otherwise `null` returned.
 */
@JvmName("toIntOrNull")
public inline fun UInt.toIntOrNull(): Int? = if (this >= Int.MAX_VALUE.asUInt()) null else this.asInt()


/**
 * Converts this [Long] value to [ULong].
 *
 * If this value is positive, the resulting [ULong] value represents the same numerical value as this [Long].
 * Otherwise `null` returned.
 */
@JvmName("toULongOrNull")
public inline fun Long.toULongOrNull(): ULong? = if (this < 0) null else this.asULong()

/**
 * Converts this [ULong] value to [Long].
 *
 * If this value is less than or equals to [Long.MAX_VALUE], the resulting [Long] value represents
 * the same numerical value as this [ULong].
 * Otherwise `null` returned.
 */
@JvmName("toLongOrNull")
public inline fun ULong.toLongOrNull(): Long? = if (this >= Long.MAX_VALUE.asULong()) null else this.asLong()


/********************************************* Conversion error formatters ****************************************/

@Suppress("FunctionName")
@JvmName("_formatUnsignedToSignedOverflow")
private inline fun <reified S, reified U> _formatUnsignedToSignedOverflow(value: U, limit: S) =
	"Can't convert ${U::class.simpleName!!} value to ${S::class.simpleName!!} because $value > $limit"

@Suppress("FunctionName")
@JvmName("_formatSignedToUnsignedNegative")
private inline fun <reified S, reified U> _formatSignedToUnsignedNegative(value: S) =
	"Can't convert ${S::class.simpleName!!} value to ${U::class.simpleName!!} because $value is negative"


@JvmName("formatToUByteError")
public fun Byte.formatToUByteError(): String = _formatSignedToUnsignedNegative<Byte, UByte>(this)

@JvmName("formatToByteError")
public fun UByte.formatToByteError(): String = _formatUnsignedToSignedOverflow<Byte, UByte>(this, Byte.MAX_VALUE)

@JvmName("formatToUShortError")
public fun Short.formatToUShortError(): String = _formatSignedToUnsignedNegative<Short, UShort>(this)

@JvmName("formatToShortError")
public fun UShort.formatToShortError(): String = _formatUnsignedToSignedOverflow<Short, UShort>(this, Short.MAX_VALUE)

@JvmName("formatToUIntError")
public fun Int.formatToUIntError(): String = _formatSignedToUnsignedNegative<Int, UInt>(this)

@JvmName("formatToIntError")
public fun UInt.formatToIntError(): String = _formatUnsignedToSignedOverflow<Int, UInt>(this, Int.MAX_VALUE)

@JvmName("formatToULongError")
public fun Long.formatToULongError(): String = _formatSignedToUnsignedNegative<Long, ULong>(this)

@JvmName("formatToLongError")
public fun ULong.formatToLongError(): String = _formatUnsignedToSignedOverflow<Long, ULong>(this, Long.MAX_VALUE)


/********************************************* Safe sign casts (or exception) ****************************************/

/**
 * Converts this [Byte] value to [UByte].
 *
 * If this value is positive, the resulting [UByte] value represents the same numerical value as this [Byte].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toUByteOrThrow")
public inline fun Byte.toUByteOrThrow(exception: (Byte) -> Nothing = { b -> throw IllegalArgumentException(b.formatToUByteError()) }): UByte {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toUByteOrNull() ?: exception(this)
}


/**
 * Converts this [UByte] value to [Byte].
 *
 * If this value is less than or equals to [Byte.MAX_VALUE], the resulting [Byte] value represents
 * the same numerical value as this [UByte].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toByteOrThrow")
public inline fun UByte.toByteOrThrow(exception: (UByte) -> Nothing = { b -> throw IllegalArgumentException(b.formatToByteError()) }): Byte {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toByteOrNull() ?: exception(this)
}

/**
 * Converts this [Short] value to [UShort].
 *
 * If this value is positive, the resulting [UShort] value represents the same numerical value as this [Short].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toUShortOrThrow")
public inline fun Short.toUShortOrThrow(exception: (Short) -> Nothing = { b -> throw IllegalArgumentException(b.formatToUShortError()) }): UShort {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toUShortOrNull() ?: exception(this)
}


/**
 * Converts this [UShort] value to [Short].
 *
 * If this value is less than or equals to [Short.MAX_VALUE], the resulting [Short] value represents
 * the same numerical value as this [UShort].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toShortOrThrow")
public inline fun UShort.toShortOrThrow(exception: (UShort) -> Nothing = { b -> throw IllegalArgumentException(b.formatToShortError()) }): Short {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toShortOrNull() ?: exception(this)
}

/**
 * Converts this [Int] value to [UInt].
 *
 * If this value is positive, the resulting [UInt] value represents the same numerical value as this [Int].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toUIntOrThrow")
public inline fun Int.toUIntOrThrow(exception: (Int) -> Nothing = { b -> throw IllegalArgumentException(b.formatToUIntError()) }): UInt {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toUIntOrNull() ?: exception(this)
}

/**
 * Converts this [UInt] value to [Int].
 *
 * If this value is less than or equals to [Int.MAX_VALUE], the resulting [Int] value represents
 * the same numerical value as this [UInt].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toIntOrThrow")
public inline fun UInt.toIntOrThrow(exception: (UInt) -> Nothing = { b -> throw IllegalArgumentException(b.formatToIntError()) }): Int {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toIntOrNull() ?: exception(this)
}


/**
 * Converts this [Long] value to [ULong].
 *
 * If this value is positive, the resulting [ULong] value represents the same numerical value as this [Long].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toULongOrThrow")
public inline fun Long.toULongOrThrow(exception: (Long) -> Nothing = { b -> throw IllegalArgumentException(b.formatToULongError()) }): ULong {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toULongOrNull() ?: exception(this)
}


/**
 * Converts this [ULong] value to [Long].
 *
 * If this value is less than or equals to [Long.MAX_VALUE], the resulting [Long] value represents
 * the same numerical value as this [ULong].
 * Otherwise [exception] function called (by default throws [IllegalArgumentException]).
 */
@JvmName("toLongOrThrow")
public inline fun ULong.toLongOrThrow(exception: (ULong) -> Nothing = { b -> throw IllegalArgumentException(b.formatToLongError()) }): Long {
	contract {
		callsInPlace(exception, InvocationKind.AT_MOST_ONCE)
	}

	return this.toLongOrNull() ?: exception(this)
}


